@@ -1,6 +1,6 @@
 # pylint: disable=unused-argument,invalid-name,line-too-long
 import logging
-from contextlib import contextmanager
+from contextlib import ExitStack, contextmanager
 from itertools import zip_longest
 from pathlib import Path
 from typing import List, Optional, Type, TypeVar
@@ -8,6 +8,7 @@ from typing import List, Optional, Type, TypeVar
 from alembic.autogenerate import comparators, renderers
 from alembic.operations import Operations
 from flupy import flu
+from sqlalchemy import exc as sqla_exc
 from sqlalchemy.orm import Session
 
 from alembic_utils.dependencies import defer_dependent
@@ -17,27 +18,87 @@ from alembic_utils.exceptions import (
     UnreachableException,
 )
 from alembic_utils.reversible_op import ReversibleOp
-from alembic_utils.statement import escape_colon, normalize_whitespace, strip_terminating_semicolon
+from alembic_utils.statement import (
+    escape_colon,
+    normalize_whitespace,
+    strip_terminating_semicolon,
+)
 
-log = logging.getLogger(__name__)
+logger = logging.getLogger(__name__)
 
 T = TypeVar("T", bound="ReplaceableEntity")
 
 
 @contextmanager
-def simulate_entity(sess: Session, entity):
-    """Creates *entities* in a transaction so postgres rendered definition
+def simulate_entity(sess: Session, entity, dependencies: Optional[List[T]] = None):
+    """Creates *entiity* in a transaction so postgres rendered definition
     can be retrieved
     """
+    if not dependencies:
+        dependencies: List[T] = []
+
     try:
         sess.begin_nested()
-        with defer_dependent(sess, entity):
-            sess.execute(entity.to_sql_statement_create_or_replace())
-            yield sess
+
+        dependency_managers = [simulate_entity(sess, x) for x in dependencies]
+
+        with ExitStack() as stack:
+            # Setup all the possible dependencies
+            for mgr in dependency_managers:
+                stack.enter_context(mgr)
+
+            with defer_dependent(sess, entity):
+                sess.execute(entity.to_sql_statement_create_or_replace())
+                yield sess
     finally:
         sess.rollback()
 
 
+def solve_resolution_order(sess: Session, entities):
+    """Solve for an entity resolution order that increases the probability that
+    a migration will suceed if, for example, two new views are created and one
+    refers to the other
+
+    This strategy will only solve for simple cases
+    """
+
+    resolved = []
+
+    # Resolve the entities with 0 dependencies first (faster)
+    logger.debug("Resolving entities with no dependencies")
+    for entity in entities:
+        try:
+            with simulate_entity(sess, entity):
+                resolved.append(entity)
+        except (sqla_exc.ProgrammingError, sqla_exc.InternalError):
+            continue
+
+    # Resolve entities with possible dependencies
+    for _ in range(len(entities)):
+        logger.info("Resolving entities with dependencies. This may take a minute")
+        n_resolved = len(resolved)
+
+        for entity in entities:
+            if entity in resolved:
+                continue
+
+            try:
+                with simulate_entity(sess, entity, dependencies=resolved):
+                    resolved.append(entity)
+            except (sqla_exc.ProgrammingError, sqla_exc.InternalError):
+                continue
+
+        if len(resolved) == n_resolved:
+            # No new entities resolved in the last iteration. Exit
+            break
+
+    for entity in entities:
+        if entity not in resolved:
+            resolved.append(entity)
+
+    return resolved
+
+
 class ReplaceableEntity:
     """A SQL Entity that can be replaced"""
 
@@ -84,9 +145,11 @@ class ReplaceableEntity:
         """ Generates a SQL "create or replace function" statement for PGFunction """
         raise NotImplementedError()
 
-    def get_database_definition(self: T, sess: Session) -> T:  # $Optional[T]:
-        """ Looks up self and return the copy existing in the database (maybe)the"""
-        with simulate_entity(sess, self) as sess:
+    def get_database_definition(
+        self: T, sess: Session, dependencies: Optional[List[T]] = None
+    ) -> T:  # $Optional[T]:
+        """Creates the entity in the database, retrieves its 'rendered' then rolls it back"""
+        with simulate_entity(sess, self, dependencies) as sess:
             # Drop self
             sess.execute(self.to_sql_statement_drop())
 
@@ -94,7 +157,7 @@ class ReplaceableEntity:
             db_entities = self.from_database(sess, schema=self.schema)
             db_entities = sorted(db_entities, key=lambda x: x.identity)
 
-        with simulate_entity(sess, self) as sess:
+        with simulate_entity(sess, self, dependencies) as sess:
             # collect all remaining entities
             all_w_self = self.from_database(sess, schema=self.schema)
             all_w_self = sorted(all_w_self, key=lambda x: x.identity)
@@ -136,13 +199,14 @@ class ReplaceableEntity:
         object_name = self.signature.split("(")[0].strip().lower()
         return f"{schema_name}_{object_name}"
 
-    def get_required_migration_op(self, sess: Session) -> Optional[ReversibleOp]:
+    def get_required_migration_op(
+        self, sess: Session, dependencies: Optional[List[T]] = None
+    ) -> Optional[ReversibleOp]:
         """Get the migration operation required for autogenerate"""
         # All entities in the database for self's schema
         entities_in_database = self.from_database(sess, schema=self.schema)
 
-        with simulate_entity(sess, self) as sess:
-            db_def = self.get_database_definition(sess)
+        db_def = self.get_database_definition(sess, dependencies=dependencies)
 
         for x in entities_in_database:
 
@@ -309,38 +373,69 @@ def register_entities(
 
         with engine.connect() as connection:
 
-            for entity in entities:
-                print(f"Processing {entity.__class__.__name__} {entity.identity}")
+            # Solve resolution order
+            try:
+                transaction = connection.begin()
+                sess = Session(bind=connection)
+                ordered_entities: List[T] = solve_resolution_order(sess, entities)
+            finally:
+                transaction.rollback()
+
+            # entities that are receiving a create or update op
+            has_create_or_update_op = []
+
+            # database rendered definitions for the entities we have a local instance for
+            # Note: used for drops
+            local_entities = []
+
+            # Required migration OPs, Create/Update/NoOp
+            for entity in ordered_entities:
+                logger.info(
+                    "Detecting required migration op %s %s",
+                    entity.__class__.__name__,
+                    entity.identity,
+                )
 
                 try:
-                    # Start a parent transaction
-                    # Bind the session within the parent transaction
                     transaction = connection.begin()
                     sess = Session(bind=connection)
-                    maybe_op = entity.get_required_migration_op(sess)
+
+                    maybe_op = entity.get_required_migration_op(
+                        sess, dependencies=has_create_or_update_op
+                    )
+
+                    local_db_def = entity.get_database_definition(
+                        sess, dependencies=has_create_or_update_op
+                    )
+                    local_entities.append(local_db_def)
+
                     if maybe_op:
                         upgrade_ops.ops.append(maybe_op)
+                        has_create_or_update_op.append(entity)
 
-                        print(
-                            f"Detected {maybe_op.__class__.__name__} for {entity.__class__.__name__} {entity.identity}"
+                        logger.info(
+                            "Detected %s op for %s %s",
+                            maybe_op.__class__.__name__,
+                            entity.__class__.__name__,
+                            entity.identity,
+                        )
+                    else:
+                        logger.info(
+                            "Detected NoOp op for %s %s",
+                            entity.__class__.__name__,
+                            entity.identity,
                         )
 
                 finally:
                     transaction.rollback()
 
+            # Required migration OPs, Drop
             try:
                 # Start a parent transaction
                 # Bind the session within the parent transaction
                 transaction = connection.begin()
                 sess = Session(bind=connection)
 
-                # Convert the user defined definitions to the database rendered definitions
-                local_entities = []
-                for ent in entities:
-                    with simulate_entity(sess, ent) as sess:
-                        local_entity = ent.get_database_definition(sess)
-                        local_entities.append(local_entity)
-
                 # All database entities currently live
                 # Check if anything needs to drop
                 for entity_class in ReplaceableEntity.__subclasses__():
@@ -358,8 +453,10 @@ def register_entities(
                             else:
                                 # No match was found locally
                                 upgrade_ops.ops.append(DropOp(db_entity))
-                                print(
-                                    f"Detected DropOp drop {db_entity.__class__.__name__} {db_entity.identity}"
+                                logger.info(
+                                    "Detected DropOp op for %s %s",
+                                    db_entity.__class__.__name__,
+                                    db_entity.identity,
                                 )
 
             finally:
